from typing import Optional

from mypy.nodes import Expression


def get_expression_fullname(expr: Expression) -> Optional[str]:
    fullname_attr = getattr(expr, "fullname", None)

    if callable(fullname_attr):
        fullname_attr = fullname_attr()

    return fullname_attr if isinstance(fullname_attr, str) else None



import os
import re
import xml.etree.ElementTree as ET
from pathlib import Path
from unittest.mock import patch

import pytest
from _pytest.capture import CaptureFixture
from mypy.errors import CompileError

from testing.util import MypyNodeFactory, WithStubTestConfig

from .analyze import analyze_stubs, compare, main
from .compare import ComparisonResult, MatchResult


class TestAnalyzeStubs(WithStubTestConfig):
    def test_analyze_missing(self, capsys: CaptureFixture) -> None:
        analyze_stubs(
            self.mypy_config_path,
            self.handwritten_stubs_path,
            self.generated_stubs_path,
            self.get_expectations_path("missing.json"),
        )

        _, err = capsys.readouterr()

        assert "missing.MISSING_CONSTANT" not in err
        assert (
            'Expected "missing.missing_function" to be '
            '"mismatch" but it was "not_found"' in err
        )
        assert (
            "Expected the following symbols to fail, but they were not processed:\n"
            " - missing.missing_decorator" in err
        )
        assert 'Symbol "missing.MissingClass" not found in generated stubs' in err
        assert (
            f"1 fail(s) were ignored, "
            f"because they were defined to be expected mismatches." in err
        )

    def test_ignore_missing_module_symbols(self, capsys: CaptureFixture) -> None:
        success = analyze_stubs(
            self.mypy_config_path,
            self.get_test_stub_path("test_ignore_missing_module_symbols"),
        )

        _, err = capsys.readouterr()

        assert 'Symbol "isort.comments.__name__" not found' not in err
        assert 'Symbol "isort.comments.__doc__" not found' not in err
        assert 'Symbol "isort.comments.__file__" not found' not in err
        assert 'Symbol "isort.comments.__package__" not found' not in err

        assert success

    def test_analyze_mismatching(self, capsys: CaptureFixture) -> None:
        analyze_stubs(
            self.mypy_config_path,
            self.handwritten_stubs_path,
            self.generated_stubs_path,
            self.get_expectations_path("mismatching.json"),
        )

        _, err = capsys.readouterr()

        assert "mismatching.mismatching_function" not in err
        assert (
            'Expected "mismatching.MISMATCHING_CONSTANT" to be "not_found" '
            'but it was "mismatch"' in err
        )
        assert "Types for mismatching.mismatch_variable do not match" in err
        assert (
            "2 fail(s) were ignored, "
            "because they were defined to be expected mismatches." in err
        )

    def test_analyze_matching(self, capsys: CaptureFixture) -> None:
        analyze_stubs(
            self.mypy_config_path,
            self.handwritten_stubs_path,
            self.generated_stubs_path,
            self.get_expectations_path("matching.json"),
        )

        _, err = capsys.readouterr()

        assert "matching.matching_function" not in err
        assert (
            'Expected "matching.MATCHING_CONSTANT" to be "not_found" but it matched'
            in err
        )
        assert (
            'Expected "matching.matching_variable" to be "mismatch" but it matched'
            in err
        )

    def test_analyze_additional_params(self, capsys: CaptureFixture) -> None:
        analyze_stubs(
            self.mypy_config_path,
            self.handwritten_stubs_path,
            self.generated_stubs_path,
            self.get_expectations_path("additional_function_params.json"),
        )

        _, err = capsys.readouterr()

        assert "functions.additional_args" not in err
        assert (
            'Expected "functions.matching_function" to be "mismatch" but it matched'
            in err
        )

    def test_analyze_checkstyle_report(self) -> None:
        report_path = "checkstyle_report.xml"
        analyze_stubs(
            self.mypy_config_path,
            self.handwritten_stubs_path,
            self.generated_stubs_path,
            None,
            report_path,
        )

        try:
            written_report = open(report_path)
            base_path = Path.resolve(
                Path(os.path.dirname(os.path.abspath(__file__))) / ".."
            )
            expected_report_path = base_path / "testing" / "test_report.xml"
            expected_report = open(expected_report_path)
            written_tree = ET.fromstring(written_report.read())
            expected_tree = ET.fromstring(
                expected_report.read().format(path=str(base_path))
            )
            assert written_tree.tag == expected_tree.tag
            assert written_tree.attrib == expected_tree.attrib
            assert len(written_tree) == len(expected_tree)
            for (written_file, expected_file) in zip(written_tree, expected_tree):
                assert written_file.tag == expected_file.tag
                assert written_file.attrib == expected_file.attrib
                assert len(written_file) == len(expected_file)
                for (written_error, expected_error) in zip(written_file, expected_file):
                    assert written_error.tag == expected_error.tag
                    assert written_error.attrib == expected_error.attrib
        finally:
            os.unlink(report_path)


class TestCompareSymbols:
    def test_generated_is_missing_a_function(self, mypy_nodes: MypyNodeFactory) -> None:
        func_def_symbol = mypy_nodes.get_missing_function_node()
        result = compare([func_def_symbol], [])

        assert all([x.match_result is MatchResult.NOT_FOUND for x in result])

    def test_generated_has_no_parameters(self, mypy_nodes: MypyNodeFactory) -> None:
        func_def, func_def_reference = mypy_nodes.get_mismatch_with_zero_parameters()
        result = compare([func_def], [func_def_reference])

        assert all([x.match_result is MatchResult.MISMATCH for x in result])

    def test_mislocated_symbol(self, mypy_nodes: MypyNodeFactory) -> None:
        mislocated_method = mypy_nodes.get_mislocated_method_handwritten()
        _, mislocated_methods_class = mypy_nodes.get_mislocated_methods_class()
        _, original_class = mypy_nodes.get_class()
        _, original_method = mypy_nodes.get_method()

        result = list(
            compare(
                [mislocated_method],
                [mislocated_methods_class, original_class, original_method],
            )
        )
        assert result == [
            ComparisonResult.create_mislocated_symbol(
                symbol=mislocated_method,
                reference=original_method,
                data={"containing_class": original_method.info},
            )
        ]


class TestCommandLineTool:
    @patch("sys.argv", ["analyze-stubs"])
    def test_missing_required_arguments(self, capsys: CaptureFixture) -> None:
        with pytest.raises(SystemExit) as ex:
            main()
        assert ex.value.code == 2

        output = capsys.readouterr().err

        assert re.search(r"analyze-stubs: error", output)
        assert re.search(
            r"the following arguments are required: -c/--config, STUBS_HANDWRITTEN",
            output,
        )

    @patch(
        "sys.argv",
        [
            "analyze-stubs",
            "-c",
            "mypy.ini",
            "testing/test-stubs/test_generated_reference_stubs",
        ],
    )
    def test_generated_reference_stubs(self, capsys: CaptureFixture) -> None:
        """
        Ensures that analyzing stubs with automatically generated reference
        stubs works as expected.
        """
        with pytest.raises(SystemExit) as ex:
            main()
        assert ex.value.code == 1

        output = capsys.readouterr().err

        # incorrectly stubbed
        assert re.search(r"Types for black.DEFAULT_LINE_LENGTH do not match", output)
        assert re.search(r"Types for black.shutdown do not match", output)
        assert re.search(
            r"Symbol \"black.NotFound\" not found in generated stubs", output
        )

        # cancel has been correctly stubbed
        assert not re.search(r"cancel", output)


class TestMislocatedSymbol(WithStubTestConfig):
    def test_mislocated_symbol(self, capsys: CaptureFixture) -> None:
        analyze_stubs(
            self.mypy_config_path,
            self.handwritten_stubs_path,
            self.generated_stubs_path,
        )

        _, err = capsys.readouterr()

        assert (
            'Found symbol "classes.ClassWithoutSuperClassInHandwritten.a_method"'
            ' in different location "classes.AClass.a_method".'
        ) in err

    def test_mislocated_symbol_expected(self, capsys: CaptureFixture) -> None:
        analyze_stubs(
            self.mypy_config_path,
            self.handwritten_stubs_path,
            self.generated_stubs_path,
            self.get_expectations_path("mislocated_symbol.json"),
        )

        _, err = capsys.readouterr()

        assert (
            'Found symbol "classes.ClassWithoutSuperClassInHandwritten.a_method"'
            ' in different location "classes.AClass.a_method".'
        ) not in err


class TestCompileError(WithStubTestConfig):
    def test_compile_error_invalid_syntax_in_base_stubs(self) -> None:
        """ Invalid stubs written by the user should cause a CompileError """
        with pytest.raises(CompileError) as ex:
            analyze_stubs(
                self.mypy_config_path,
                self.get_test_stub_path("test_compile_error_invalid_syntax"),
            )

        assert "poolmanager.pyi:7: error: invalid syntax" in ex.value.messages[0]

    def test_compile_error_invalid_syntax_in_reference(self) -> None:
        """
        Invalid reference stubs should cause a CompileError, too. This simulates the
        case where stubgen generates invalid stubs due to an internal error.
        """
        with pytest.raises(CompileError) as ex:
            analyze_stubs(
                self.mypy_config_path,
                self.handwritten_stubs_path,
                self.get_test_stub_path("test_compile_error_invalid_syntax"),
            )

        assert "poolmanager.pyi:7: error: invalid syntax" in ex.value.messages[0]



from typing import Dict, cast

from mypy.nodes import Decorator, SymbolNode, TypeAlias, TypeInfo
from mypy.types import CallableType, FuncDef, Instance, UnboundType, UnionType

from testing.util import WithStubTestConfig

from .collect import get_stub_types


class WithStubSymbols(WithStubTestConfig):
    _symbols: Dict[str, SymbolNode]

    @classmethod
    def setup_class(cls) -> None:
        stubs_path = cls.get_generated_stubs_path()
        conf_path = cls.get_mypy_config_path()

        cls._symbols = {
            s.fullname(): s
            for (s, _) in get_stub_types(str(stubs_path), str(conf_path))
        }

    @property
    def symbols(self) -> Dict[str, SymbolNode]:
        return self._symbols


class TestGetStubTypes(WithStubSymbols):
    def test_method_stub(self) -> None:
        method = cast(
            FuncDef,
            self.symbols[
                "classes.ClassWithInvalidCustomStub.return_type_less_specific"
            ],
        )

        # function arguments
        assert method.arg_names == ["self", "foo", "bar"]
        assert method.arguments[0].type_annotation is None

        int_param = cast(UnboundType, method.arguments[1].type_annotation)
        assert isinstance(int_param, UnboundType)
        assert int_param.name == "int"

        # function return value
        assert isinstance(method.type, CallableType)
        assert isinstance(method.type.ret_type, Instance)
        assert method.type.ret_type.type.name() == "bool"

    def test_type_alias(self) -> None:
        type_alias = cast(TypeAlias, self.symbols["type_aliases.Strint"])
        assert isinstance(type_alias, TypeAlias)
        assert type_alias.name() == "Strint"
        assert isinstance(type_alias.target, UnionType)

        assert str(type_alias.target.items[0]) == "builtins.str"
        assert str(type_alias.target.items[1]) == "builtins.int"

    def test_inheritance(self) -> None:
        subclass = self.symbols["classes.SubClassOfAClass"]
        assert subclass
        assert isinstance(subclass, TypeInfo)
        assert {b.type.fullname() for b in subclass.bases} == {"classes.AClass"}

    def test_inherited_definitions(self) -> None:
        another_class = self.symbols.get("classes.SubClassOfAClass")
        assert isinstance(another_class, TypeInfo)

        # the hash method is not declared in the bcrypt class directly
        assert not another_class.names.get("a_classmethod")
        assert "classes.SubClassOfAClass.a_classmethod" not in self.symbols

        # but it can be resolved from definitions inherited from base classes
        a_classmethod = another_class.get("a_classmethod")
        assert a_classmethod and isinstance(a_classmethod.node, Decorator)

    def test_base_classes(self) -> None:
        subClassOfAClass = self.symbols.get("classes.SubClassOfAClass")
        assert isinstance(subClassOfAClass, TypeInfo)
        assert {b.type.fullname() for b in subClassOfAClass.bases} == {"classes.AClass"}

        aClass = self.symbols.get("classes.AClass")
        assert isinstance(aClass, TypeInfo)
        assert {b.type.fullname() for b in aClass.bases} == {"builtins.object"}



from typing import Union

from mypy.nodes import (
    Decorator,
    FuncDef,
    OverloadedFuncDef,
    TypeAlias,
    TypeInfo,
    TypeVarExpr,
    Var,
)

RelevantSymbolNode = Union[
    Decorator, FuncDef, OverloadedFuncDef, Var, TypeInfo, TypeVarExpr, TypeAlias
]

"""
SymbolNodes that are interesting for stub files.
Excludes MypyFiles (the modules themselves), imported names and Mypy placeholders
"""



from .collect import get_stub_types
from .compare import ComparisonResult, MatchResult, compare_symbols
from .lookup import lookup_symbol
from .types import RelevantSymbolNode

__all__ = [
    "compare_symbols",
    "ComparisonResult",
    "get_stub_types",
    "lookup_symbol",
    "MatchResult",
    "RelevantSymbolNode",
]



from unittest.mock import Mock

from mypy.nodes import Expression

from .utils import get_expression_fullname


class TestUtils:
    def test_get_expression_fullname_when_attr_is_str(self) -> None:
        expected_fullname = "random.expression.fullname"
        expression = Mock(fullname=expected_fullname, spec=Expression)
        assert get_expression_fullname(expression) == expected_fullname

    def test_get_expression_fullname_when_attr_is_none(self) -> None:
        expression = Mock(fullname=None, spec=Expression)
        assert get_expression_fullname(expression) is None

    def test_get_expression_fullname_when_attr_is_not_str_nor_none(self) -> None:
        expression = Mock(fullname=123, spec=Expression)
        assert get_expression_fullname(expression) is None

    def test_get_expression_fullname_when_attr_does_not_exist(self) -> None:
        expression = Mock(spec=Expression)
        assert get_expression_fullname(expression) is None

    def test_get_expression_fullname_when_attr_is_function_and_return_str(self) -> None:
        expected_fullname = "random.expression.fullname.from_method"
        fullname_mock = Mock(return_value=expected_fullname)
        expression = Mock(fullname=fullname_mock, spec=Expression)
        assert get_expression_fullname(expression) is expected_fullname

    def test_get_expression_fullname_when_attr_is_function_and_do_not_return_str(
        self,
    ) -> None:
        fullname_mock = Mock(return_value=13245)
        expression = Mock(fullname=fullname_mock, spec=Expression)
        assert get_expression_fullname(expression) is None



import pytest
from mypy.nodes import Var
from mypy.types import NoneType

from testing.util import MypyNodeFactory

from .compare import ComparisonResult, MatchResult, compare_symbols


def assert_match(result: ComparisonResult) -> None:
    assert result.match_result is MatchResult.MATCH


def assert_mismatch(result: ComparisonResult) -> None:
    assert result.match_result is MatchResult.MISMATCH


class TestMatchResult:
    def test_declare_mismatch(self) -> None:
        assert MatchResult.declare_mismatch("not_found") is MatchResult.NOT_FOUND
        assert MatchResult.declare_mismatch("mismatch") is MatchResult.MISMATCH
        assert (
            MatchResult.declare_mismatch("mislocated_symbol")
            is MatchResult.MISLOCATED_SYMBOL
        )
        with pytest.raises(ValueError, match=r".*not a valid mismatch type.*"):
            MatchResult.declare_mismatch("match")
        with pytest.raises(ValueError, match=r".*not a valid mismatch type.*"):
            MatchResult.declare_mismatch("blablabla")


class TestComparisonResult:
    def test_create_not_found_does_not_match(self) -> None:
        own_symbol = Var("x")
        data = {"fullname": "package.x"}
        result = ComparisonResult.create_not_found(own_symbol, data=data)
        assert result.match_result is MatchResult.NOT_FOUND
        assert result.symbol is own_symbol
        assert result.reference is None
        assert result.message is not None
        assert result.data is data

    def test_create_mismatch(self) -> None:
        own_symbol = Var("x", NoneType())
        other_symbol = Var("x", None)
        data = {"name_a": "package.x", "name_b": "package.x"}
        result = ComparisonResult.create_mismatch(own_symbol, other_symbol, data=data)
        assert result.match_result is MatchResult.MISMATCH
        assert result.symbol is own_symbol
        assert result.reference is other_symbol
        assert result.message is not None
        assert result.data is data

    def test_create_match(self) -> None:
        own_symbol = Var("x", NoneType())
        other_symbol = Var("x", NoneType())
        data = {"name_a": "package.x", "name_b": "package.x"}
        result = ComparisonResult.create_match(own_symbol, other_symbol, data=data)
        assert result.match_result is MatchResult.MATCH
        assert result.symbol is own_symbol
        assert result.reference is other_symbol
        assert result.message is not None
        assert result.data is data


class TestCompareNone:
    def test_right_type_is_none_succeeds(self) -> None:
        own = Var("x", NoneType())  # the type of "None"
        other = Var("x", None)  # actually missing
        result = compare_symbols(own, other)
        assert result.match_result is MatchResult.MATCH
        assert result.message == "Generated type is None"

    def test_both_types_are_none_succeeds(self) -> None:
        own = Var("x", None)
        other = Var("x", None)
        result = compare_symbols(own, other)
        assert result.match_result is MatchResult.MATCH

    def test_left_is_none_fails(self) -> None:
        own = Var("x", None)  # actually missing
        other = Var("x", NoneType())  # the type of "None"
        result = compare_symbols(own, other)
        assert result.match_result is MatchResult.MISMATCH

    def test_both_types_are_none_type_succeeds(self) -> None:
        own = Var("x", NoneType())
        other = Var("x", NoneType())
        result = compare_symbols(own, other)
        assert result.match_result is MatchResult.MATCH


class TestCompareTypeInfos:
    def test_type_infos_with_same_name_match(self, mypy_nodes: MypyNodeFactory) -> None:
        cls, cls_reference = mypy_nodes.get_class()
        result = compare_symbols(cls, cls_reference)

        assert result.match_result is MatchResult.MATCH

    def test_type_infos_with_different_names_mismatch(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        a_cls, _ = mypy_nodes.get_class()
        antoher_cls, _ = mypy_nodes.get_another_class()
        result = compare_symbols(a_cls, antoher_cls)

        assert result.match_result is MatchResult.MISMATCH


class TestComparePrimitiveVars:
    def test_bool_more_specific_than_int(self, mypy_nodes: MypyNodeFactory) -> None:
        int_var = mypy_nodes.get_int_var()
        bool_var = mypy_nodes.get_bool_var()

        result_a = compare_symbols(int_var, bool_var)
        result_b = compare_symbols(bool_var, int_var)

        assert result_a.match_result is MatchResult.MISMATCH
        assert result_b.match_result is MatchResult.MATCH

    def test_str_and_int_dont_match(self, mypy_nodes: MypyNodeFactory) -> None:
        int_var = mypy_nodes.get_int_var()
        str_var = mypy_nodes.get_str_var()

        result_a = compare_symbols(int_var, str_var)
        result_b = compare_symbols(str_var, int_var)

        assert result_a.match_result is MatchResult.MISMATCH
        assert result_b.match_result is MatchResult.MISMATCH


class TestCompareMethods:
    def test_compare_method(self, mypy_nodes: MypyNodeFactory) -> None:
        meth, meth_ref = mypy_nodes.get_method()
        result = compare_symbols(meth, meth_ref)
        assert result.match_result is MatchResult.MATCH

    def test_compare_classmethod(self, mypy_nodes: MypyNodeFactory) -> None:
        meth, meth_ref = mypy_nodes.get_classmethod()
        result = compare_symbols(meth, meth_ref)
        assert result.match_result is MatchResult.MATCH

    def test_compare_overridden_method(self, mypy_nodes: MypyNodeFactory) -> None:
        override, override_meth = mypy_nodes.get_overridden_method()
        result = compare_symbols(override, override_meth)
        assert result.match_result is MatchResult.MATCH

    def test_compare_overridden_classmethod(self, mypy_nodes: MypyNodeFactory) -> None:
        override, override_meth = mypy_nodes.get_overridden_classmethod()
        result = compare_symbols(override, override_meth)
        assert result.match_result is MatchResult.MATCH

    def test_argument_order_wrong(self, mypy_nodes: MypyNodeFactory) -> None:
        meth, meth_ref = mypy_nodes.get_argument_order_wrong()
        result = compare_symbols(meth, meth_ref)

        assert result.match_result is MatchResult.MISMATCH

    def test_argument_names_wrong_does_mismatch(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        meth, meth_ref = mypy_nodes.get_argument_names_wrong()
        result = compare_symbols(meth, meth_ref)
        assert result.match_result is MatchResult.MISMATCH

    def test_argument_types_wrong(self, mypy_nodes: MypyNodeFactory) -> None:
        meth, meth_ref = mypy_nodes.get_argument_types_wrong()
        result = compare_symbols(meth, meth_ref)
        assert result.match_result is MatchResult.MISMATCH

    def test_return_type_less_specific_does_match(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        meth, meth_ref = mypy_nodes.get_return_type_less_specific()
        result = compare_symbols(meth, meth_ref)
        assert result.match_result is MatchResult.MATCH

    def test_return_type_wrong(self, mypy_nodes: MypyNodeFactory) -> None:
        meth, meth_ref = mypy_nodes.get_return_type_wrong()
        result = compare_symbols(meth, meth_ref)
        assert result.match_result is MatchResult.MISMATCH


class TestCompareFunctions:
    def test_matching_functions(self, mypy_nodes: MypyNodeFactory) -> None:
        func_def, func_ref = mypy_nodes.get_matching_func_node()
        result = compare_symbols(func_def, func_ref)
        assert_match(result)

    def test_func_def_mismatches_when_handwritten_stub_has_additional_optional_args(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        func_def, func_def_reference = mypy_nodes.get_additional_optional_args_node()
        result = compare_symbols(func_def, func_def_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_func_def_mismatches_when_handwritten_stub_has_additional_args(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        func_def, func_def_reference = mypy_nodes.get_additional_args_node()
        result = compare_symbols(func_def, func_def_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_matching_with_arg_star(self, mypy_nodes: MypyNodeFactory) -> None:
        func, func_ref = mypy_nodes.get_matching_with_arg_star()
        result = compare_symbols(func, func_ref)
        assert_match(result)

    def test_matching_with_missing_arg_star(self, mypy_nodes: MypyNodeFactory) -> None:
        func, func_ref = mypy_nodes.get_matching_with_missing_arg_star()
        result = compare_symbols(func, func_ref)
        assert_mismatch(result)

    def test_mismatching_with_arg_star(self, mypy_nodes: MypyNodeFactory) -> None:
        func, func_ref = mypy_nodes.get_mismatching_with_arg_star()
        result = compare_symbols(func, func_ref)
        assert_mismatch(result)

    def test_mismatching_with_additional_arg_star(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        func, func_ref = mypy_nodes.get_mismatching_with_additional_arg_star()
        result = compare_symbols(func, func_ref)
        assert_mismatch(result)

    def test_matching_with_kwarg_star2(self, mypy_nodes: MypyNodeFactory) -> None:
        func, func_ref = mypy_nodes.get_matching_with_kwarg_star2()
        result = compare_symbols(func, func_ref)
        assert_match(result)

    def test_matching_with_missing_kwarg_star2(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        func, func_ref = mypy_nodes.get_matching_with_missing_kwarg_star2()
        result = compare_symbols(func, func_ref)
        assert_mismatch(result)

    def test_mismatching_with_kwarg_star2(self, mypy_nodes: MypyNodeFactory) -> None:
        func, func_ref = mypy_nodes.get_mismatching_with_kwarg_star2()
        result = compare_symbols(func, func_ref)
        assert_mismatch(result)

    def test_mismatching_with_additional_kwarg_star2(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        func, func_ref = mypy_nodes.get_mismatching_with_additional_kwarg_star2()
        result = compare_symbols(func, func_ref)
        assert_mismatch(result)

    def test_overload_mismatches_if_any_handwritten_stub_has_additional_args(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        (
            overloaded_def,
            overloaded_reference,
        ) = mypy_nodes.get_overloaded_additional_args_node()
        result = compare_symbols(overloaded_def, overloaded_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_overload_mismatch_if_any_handwritten_stub_has_additional_optional_args(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        (
            overloaded_def,
            overloaded_reference,
        ) = mypy_nodes.get_overloaded_additional_optional_args_node()
        result = compare_symbols(overloaded_def, overloaded_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_decorated_function_matches(self, mypy_nodes: MypyNodeFactory) -> None:
        decorated, decorated_reference = mypy_nodes.get_decorated_function()
        result = compare_symbols(decorated, decorated_reference)
        assert result.match_result is MatchResult.MATCH

    def test_decorated_function_mismatches_when_handwritten_stub_has_additional_args(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        decorated, decorated_reference = mypy_nodes.get_decorated_with_additional_args()
        result = compare_symbols(decorated, decorated_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_decorated_mismatches_if_handwritten_stub_has_additional_optional_args(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        (
            decorated,
            decorated_reference,
        ) = mypy_nodes.get_decorated_with_additional_optional_args()
        result = compare_symbols(decorated, decorated_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_generated_has_no_parameters_and_return_type(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        (
            func_def,
            func_def_reference,
        ) = mypy_nodes.get_no_parameters_and_return_type_node()
        result = compare_symbols(func_def, func_def_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_generated_has_no_parameters(self, mypy_nodes: MypyNodeFactory) -> None:
        func_def, func_def_reference = mypy_nodes.get_mismatch_with_zero_parameters()
        result = compare_symbols(func_def, func_def_reference)
        assert result.match_result is MatchResult.MISMATCH

    def test_no_annotation_matches_if_argc_matches(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        func_def, func_def_reference = mypy_nodes.get_function_with_no_annotation()
        result = compare_symbols(func_def, func_def_reference)
        assert result.match_result is MatchResult.MATCH

    def test_args_with_no_annotation_matches_if_argc_matches(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        (
            func_def,
            func_def_reference,
        ) = mypy_nodes.get_function_with_args_but_no_annotation()
        result = compare_symbols(func_def, func_def_reference)
        assert result.match_result is MatchResult.MATCH



"""
API for analyzing Python stubs using mypy.
"""
import os
from os.path import abspath
from typing import Iterable, Optional, Set, Tuple

from mypy.build import BuildResult, State, build
from mypy.main import process_options
from mypy.nodes import (
    GDEF,
    MDEF,
    Decorator,
    FuncDef,
    MypyFile,
    OverloadedFuncDef,
    SymbolNode,
    TypeAlias,
    TypeInfo,
    TypeVarExpr,
    Var,
)

from .types import RelevantSymbolNode

IGNORED_MODULE_SYMBOLS = [
    # because docstrings are not generated by stubgen
    "__doc__",
    # created by the import system, do not need to be handwritten:
    "__name__",
    "__package__",
    "__file__",
]
"""
Module level definitions that will not be collected by collect_types.
"""


def _mypy_analyze(
    mypy_conf_path: str, root_path: str, stubs_path: Optional[str] = None
) -> BuildResult:
    """
    Parse and analyze the types of the code in root_path.

    :param mypy_conf_path: path to a mypy.ini
    :param root_path: path to the code directory where the type analysis is started
    :param stubs_path: path to the directory of stubs for mypy to use
    :returns: Mypy's analysis result
    """
    # The call to `build.build` is inspired by `mypy/mypy/main.py::main`
    # `build` is not a documented public API
    args = [
        "--config-file",
        mypy_conf_path,
        "--no-incremental",
        "--cache-dir=" + ("nul" if os.name == "nt" else "/dev/null"),
        root_path,
    ]
    sources, options = process_options(args)
    if stubs_path is not None:
        options = options.apply_changes({"mypy_path": [stubs_path]})

    return build(sources, options, None, None)


def is_stubbed_module(module: State) -> bool:
    """
    Check if a module's types were loaded from a stub.

    :param module: The module to check
    """
    return module.path is not None and module.path.endswith(".pyi")


def collect_types(
    symbol_node: SymbolNode, collected_types: Optional[Set[str]] = None
) -> Iterable[RelevantSymbolNode]:
    """
    Collect all relevant type definitions of the symbols in the given node.

    :param symbol_node: any symbol node, e.g. MypyFile (BuildResult.graph.tree)
    :param collected_types: used to avoid collecting duplicates
    """
    if not collected_types:
        collected_types = set()

    # ignore builtins because we don't provide stubs for them
    if "builtins" in symbol_node.fullname():
        return

    # do not collect types twice
    if symbol_node.fullname() in collected_types:
        return
    collected_types.add(symbol_node.fullname())

    if isinstance(symbol_node, MypyFile):
        # the symbol node represents a Python module
        for symbol in symbol_node.names.values():
            # only global and class member definitions are interesting
            if symbol.kind not in [GDEF, MDEF]:
                continue

            if symbol.node and symbol.node.name() in IGNORED_MODULE_SYMBOLS:
                continue

            if symbol.node and symbol.module_public:
                yield from collect_types(symbol.node, collected_types)
    elif isinstance(symbol_node, TypeInfo):
        # the symbol represents a class definition
        yield symbol_node
        for class_member in symbol_node.names.values():
            if class_member.node:
                yield from collect_types(class_member.node, collected_types)
    elif isinstance(
        symbol_node,
        (Decorator, FuncDef, OverloadedFuncDef, Var, TypeAlias, TypeVarExpr),
    ):
        # the symbol represents a function definition,
        # variable, type alias or generic TypeVar
        yield symbol_node
    else:
        assert False, f"Unexpected symbol type {type(symbol_node)}"


def get_stub_types(
    stubs_path: str, mypy_conf_path: str, root_path: Optional[str] = None
) -> Iterable[Tuple[RelevantSymbolNode, str]]:
    """
    Analyze the stub files in stubs_path and return module
    and class definitions of stubs as symbol nodes.

    Only relevant symbol nodes (e.g. for variables, functions, classes, methods)
    are returned.
    They contain the type annotation information.

    :param stubs_path: where all the stub files are located
    :param mypy_conf_path: path to mypy.ini
    :param root_path: path to the code directory where the type analysis is started
    """
    stubs_path = abspath(stubs_path)

    if root_path:
        build_result = _mypy_analyze(mypy_conf_path, root_path, stubs_path)
    else:
        build_result = _mypy_analyze(mypy_conf_path, stubs_path)

    stubbed_modules = {
        module
        for module in build_result.graph.values()
        if module.path
        and is_stubbed_module(module)
        and module.path.startswith(stubs_path)
    }

    for module in stubbed_modules:
        if module.tree:
            assert module.path
            yield from (
                (stub_type, module.path) for stub_type in collect_types(module.tree)
            )



from copy import copy

from mypy.nodes import FUNC_NO_INFO

from testing.util import MypyNodeFactory

from .lookup import LookupResult, lookup_symbol


class TestLookupOfMislocatedSymbols:
    def test_mislocated_symbol(self, mypy_nodes: MypyNodeFactory) -> None:
        mislocated_handwritten = mypy_nodes.get_mislocated_method_handwritten()
        gen_map = mypy_nodes.get_generated_stubs_map()
        mislocated_actual_location = (
            mypy_nodes.get_mislocated_method_actual_location_generated()
        )

        result = lookup_symbol(gen_map, mislocated_handwritten)
        assert result == LookupResult(
            symbol=mislocated_actual_location,
            containing_class=mislocated_actual_location.info,
        )

    def test_mislocated_symbol_without_cls(self, mypy_nodes: MypyNodeFactory) -> None:
        mislocated_handwritten = mypy_nodes.get_mislocated_method_handwritten()
        gen_map = mypy_nodes.get_generated_stubs_map()

        original_func_info = mislocated_handwritten.info
        mislocated_handwritten.info = FUNC_NO_INFO

        result = lookup_symbol(gen_map, mislocated_handwritten)
        assert result == LookupResult(symbol=None, containing_class=None)

        mislocated_handwritten.info = original_func_info

    def test_mislocated_symbol_with_cls_not_found(
        self, mypy_nodes: MypyNodeFactory
    ) -> None:
        mislocated_handwritten = mypy_nodes.get_mislocated_method_handwritten()
        gen_map = copy(mypy_nodes.get_generated_stubs_map())

        del gen_map[mislocated_handwritten.info.fullname()]
        result = lookup_symbol(gen_map, mislocated_handwritten)
        assert result == LookupResult(symbol=None, containing_class=None)

    def test_correctly_placed_symbol(self, mypy_nodes: MypyNodeFactory) -> None:
        hand, gen = mypy_nodes.get_method()
        gen_map = mypy_nodes.get_generated_stubs_map()

        result = lookup_symbol(gen_map, hand)
        assert result == LookupResult(symbol=gen, containing_class=gen.info)



import pytest
from mypy.nodes import TypeVarExpr

from testing.util import MypyNodeFactory

from .compare import MatchResult, compare_symbols


class TestCompareTypevars:
    @pytest.mark.parametrize(  # type: ignore
        "node_name, expected_result",
        [
            ("typevars.PlainTypeVar", MatchResult.MATCH),
            ("typevars.BoundTypeVar", MatchResult.MATCH),
            ("typevars.BoundTypeVarMoreSpecific", MatchResult.MATCH),
            ("typevars.ValuesTypeVar", MatchResult.MATCH),
            ("typevars.ValuesTypeVarWrongOrder", MatchResult.MISMATCH),
            ("typevars.ValuesTypeVarNoValuesInGenerated", MatchResult.MISMATCH),
            ("typevars.ValuesTypeVarMoreSpecific", MatchResult.MATCH),
        ],
    )
    def test_compare_typevars(
        self, node_name: str, expected_result: MatchResult, mypy_nodes: MypyNodeFactory
    ) -> None:
        symbol, reference = mypy_nodes.get(node_name, TypeVarExpr)
        result = compare_symbols(symbol, reference)
        assert result.match_result is expected_result



from typing import Dict, NamedTuple, Optional

from mypy.nodes import FUNC_NO_INFO, SymbolNode, TypeInfo

from .types import RelevantSymbolNode


class LookupResult(NamedTuple):
    symbol: Optional[SymbolNode]
    """Symbol that was found"""
    containing_class: Optional[TypeInfo]
    """Class the symbol was found on"""


def get_symbol_class(symbol: SymbolNode) -> Optional[TypeInfo]:
    """
    Get the class the given symbol is defined on.

    :param symbol: Symbol to retrieve the class of
    """
    cls = getattr(symbol, "info", None)
    if cls and isinstance(cls, TypeInfo):
        return cls

    return None


def lookup_symbol(
    symbol_map: Dict[str, RelevantSymbolNode], symbol_to_lookup: SymbolNode
) -> LookupResult:
    """
    Find the given symbol in the symbol map.

    Ideally the symbol is just found under the exact same ``fullname``.
    If not and the symbol is a method, this will take the symbol's class and look it up
    in the map and then try to resolve the symbol on that class using its method
    resolution order.

    If the result of the lookup has a different ``fullname`` than the original symbol
    the given symbol is defined at a different point in the class hierarchy than
    expected.

    :param symbol_map: Dictionary for looking up symbols by their full name
    :param symbol_to_lookup: Symbol to search for
    :return: The found symbol (if any) and the class it was found on (if any)
    """
    fail = LookupResult(None, None)

    symbol = symbol_map.get(symbol_to_lookup.fullname())
    if symbol:
        return LookupResult(symbol, get_symbol_class(symbol))

    # Check if we have a class on the symbol we're looking up
    cls_to_lookup = getattr(symbol_to_lookup, "info", None)
    if not cls_to_lookup or cls_to_lookup == FUNC_NO_INFO:
        return fail

    symbol_cls = symbol_map.get(cls_to_lookup.fullname())

    if not symbol_cls or not isinstance(symbol_cls, TypeInfo):
        return fail

    found_symbol_table_node = symbol_cls.get(symbol_to_lookup.name())
    if not found_symbol_table_node or not found_symbol_table_node.node:
        return fail

    return LookupResult(
        found_symbol_table_node.node, get_symbol_class(found_symbol_table_node.node)
    )



import re
import sys
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from collections import defaultdict
from enum import Enum
from importlib.util import find_spec
from json import loads as json_loads
from json.decoder import JSONDecodeError
from os import linesep, scandir
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from traceback import format_exception
from typing import (
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
)
from xml.etree.ElementTree import Element, ElementTree, SubElement

from mypy.nodes import TypeAlias, TypeVarExpr, Var
from mypy.stubgen import generate_stubs, parse_options
from schema import Or, Schema, SchemaError, Use

from .collect import get_stub_types
from .compare import ComparisonResult, MatchResult, compare_symbols
from .lookup import lookup_symbol
from .types import RelevantSymbolNode

EXPECTED_MISMATCH_SCHEMA = Schema(Or({}, {str: Use(MatchResult.declare_mismatch)}))
CHECK_FILE_ERROR = 'Check "{file_path}" to fix.'
MATCH_FOUND_ERROR = (
    'Expected "{symbol}" to be "{mismatch_type}" but it matched.'
    f"{linesep}{CHECK_FILE_ERROR}"
)
WRONG_MISMATCH_ERROR = (
    'Expected "{symbol}" to be "{expected}" but it was "{received}".'
    f"{linesep}{CHECK_FILE_ERROR}"
)
UNUSED_DEFINITION_ERROR = (
    "Expected the following symbols to fail, "
    f"but they were not processed:{linesep}"
    "{symbols}"
    f"{linesep}{CHECK_FILE_ERROR}"
)
FILE_NOT_FOUND_WARNING = (
    'WARNING: Provided file for expected mismatches ("{file_path}") not found.'
)
SUCCESS_MESSAGE = "Successfully validated {total} stubs."
FAIL_MESSAGE = "Failure: {failed} of {total} stubs seem not to be valid."
IGNORE_MESSAGE = (
    "{ignored} fail(s) were ignored, "
    "because they were defined to be expected mismatches."
)


class EvaluationResult(Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    EXPECTED_FAILURE = "expected_failure"


def write_error(
    *messages: str, sep: str = "", symbol: Optional[RelevantSymbolNode] = None
) -> None:
    sys.stderr.write(sep.join(messages))
    sys.stderr.write(linesep)


def parse_command_line() -> Namespace:
    parser = ArgumentParser(
        description=dedent(
            """\
        Analyze a set of (handcrafted) mypy stubs by comparing them to (generated)
        reference stubs
        """
        ),
        formatter_class=RawTextHelpFormatter,
    )
    required_group = parser.add_argument_group("required arguments")
    required_group.add_argument(
        "-c", "--config", required=True, help="Mypy config file"
    )
    parser.add_argument(
        "-e",
        "--expected-mismatches",
        required=False,
        default=None,
        help=dedent(
            """\
        A JSON file, which defines expected mismatching
        symbols and their match results. If any symbol is
        declared in an expected_mismatches JSON file,
        %(prog)s will count it as an expected failure, and
        ignore this inconsistency.

        Example contents:
        {
            "my.module.function: "mismatch",
            "another.module.Class: "not_found"
        }

        According to the example above, we expect the signature
        of my.module.function to mismatch, and module.Class to
        be missing in the generated stubs. %(prog)s will
        ignore these inconsistencies.
        """
        ),
    )
    parser.add_argument(
        dest="stubs_handwritten",
        metavar="STUBS_HANDWRITTEN",
        help=dedent(
            """\
        Directory of handwritten stubs that need to be
        analyzed
        """
        ),
    )
    parser.add_argument(
        "-r",
        "--reference",
        required=False,
        default=None,
        metavar="REFERENCE_STUBS",
        help=dedent(
            """
        Directory of reference stubs to compare against. If
        not specified stubgen will be used to generate the
        reference stubs.
        """
        ),
    )
    parser.add_argument(
        "-x",
        "--checkstyle-report",
        required=False,
        default=None,
        help=dedent(
            """
        Write an xml report in checkstyle format to the given file.
        """
        ),
    )
    return parser.parse_args()


def compare(
    hand_written: Iterable[RelevantSymbolNode], generated: Iterable[RelevantSymbolNode]
) -> Generator[ComparisonResult, None, None]:
    """Compare hand written to generated stubs."""
    gen_map: Dict[str, RelevantSymbolNode] = {
        symbol.fullname(): symbol for symbol in generated
    }

    for symbol in hand_written:
        name = symbol.fullname()
        if name in gen_map:
            yield compare_symbols(symbol, gen_map[name])
        elif isinstance(symbol, (TypeAlias, TypeVarExpr, Var)) and re.match(
            r"_[^_].*", name.split(".")[-1]
        ):
            # Ignore symbols that begin with (exactly) one _,
            # since we assume they are private
            continue
        else:
            lookup_result = lookup_symbol(gen_map, symbol)
            generated_symbol = lookup_result.symbol
            if generated_symbol:
                yield ComparisonResult.create_mislocated_symbol(
                    symbol=symbol,
                    reference=generated_symbol,
                    data={"containing_class": lookup_result.containing_class},
                )
            else:
                yield ComparisonResult.create_not_found(symbol)


def setup_expected_mismatches(
    file_path: Optional[str] = None,
) -> Tuple[Dict[str, MatchResult], Set[str]]:
    if not file_path:
        return dict(), set()

    mismatches: Dict[str, MatchResult] = {}
    mismatches_file = Path(file_path)
    if not mismatches_file.exists():
        write_error(FILE_NOT_FOUND_WARNING.format(file_path=file_path))
        return dict(), set()
    mismatches = EXPECTED_MISMATCH_SCHEMA.validate(
        json_loads(mismatches_file.read_text())
    )
    unused_mismatches = set(mismatches.keys())
    return mismatches, unused_mismatches


def evaluate_compare_result(
    compare_result: ComparisonResult,
    mismatches: Dict[str, MatchResult],
    mismatches_left: Set[str],
    expected_mismatches_path: Optional[str] = None,
    *,
    loggers: List[Callable[..., None]],
) -> EvaluationResult:
    symbol = compare_result.symbol
    symbol_name = compare_result.symbol_name
    match_result = compare_result.match_result
    evaluation_result = EvaluationResult.SUCCESS
    expected_mismatch = mismatches.get(symbol_name)

    if expected_mismatch is None:
        if match_result is not MatchResult.MATCH:
            evaluation_result = EvaluationResult.FAILURE
            for logger in loggers:
                logger(
                    linesep, compare_result.message, symbol=symbol,
                )
    else:
        mismatches_left.remove(symbol_name)
        if match_result is MatchResult.MATCH:
            evaluation_result = EvaluationResult.FAILURE
            assert expected_mismatches_path
            for logger in loggers:
                logger(
                    linesep,
                    MATCH_FOUND_ERROR.format(
                        symbol=symbol_name,
                        mismatch_type=mismatches[symbol_name].value,
                        file_path=expected_mismatches_path,
                    ),
                    symbol=symbol,
                )
        elif match_result is not expected_mismatch:
            evaluation_result = EvaluationResult.FAILURE
            assert expected_mismatches_path
            for logger in loggers:
                logger(
                    linesep,
                    WRONG_MISMATCH_ERROR.format(
                        symbol=symbol_name,
                        expected=expected_mismatch.value,
                        received=match_result.value,
                        file_path=expected_mismatches_path,
                    ),
                    symbol=symbol,
                )
        else:
            evaluation_result = EvaluationResult.EXPECTED_FAILURE
    return evaluation_result


def call_stubgen(command_line_args: List[str]) -> None:
    """
    Call stubgen like the command line tool.

    :param command_line_args: list of command line args
    """

    generate_stubs(parse_options(command_line_args), quiet=True)


def generate_stub_types(
    base_stubs_path: str, mypy_conf_path: str
) -> Iterable[Tuple[RelevantSymbolNode, str]]:
    """
    Use stubgen to generate reference stub types of the modules stubbed in
    base_stubs_path. For this to work the modules need to be installed.

    :param base_stubs_path: path to directory with (handwritten) stubs
    :param mypy_conf_path: path to mypy.ini
    :return: returns the reference stub types
    """
    with TemporaryDirectory() as reference_stubs_path:
        packages = [
            entry.name.replace(".pyi", "")
            for entry in scandir(base_stubs_path)
            if entry.is_dir() or entry.name.endswith(".pyi")
        ]
        for package in packages:
            if find_spec(package) is None:
                print(
                    f'Error: The package "{package}" is not installed. Therefore no '
                    f"reference stubs can be generated for it automatically. Use the "
                    f"option -r to provide the reference stubs manually, or install "
                    f"the package."
                )
                sys.exit(1)
            try:
                call_stubgen(
                    ["--ignore-errors", "-p", package, "-o", reference_stubs_path]
                )
            except Exception as ex:
                write_error(
                    f'Error: Generating stubs for the package "{package}" failed:',
                    linesep,
                    *format_exception(type(ex), ex, ex.__traceback__),
                )

        return list(get_stub_types(reference_stubs_path, mypy_conf_path))


class ErrorEntry(NamedTuple):
    symbol: RelevantSymbolNode
    message: str


class CheckStyleWriter:
    def __init__(self, path_map: Dict[RelevantSymbolNode, str]):
        self.path_map = path_map
        self.errors_by_file: Dict[str, List[ErrorEntry]] = defaultdict(list)

    def collect_error(
        self, *messages: str, sep: str = "", symbol: RelevantSymbolNode
    ) -> None:
        message = sep.join(messages)
        path = self.path_map[symbol]
        self.errors_by_file[path].append(ErrorEntry(symbol=symbol, message=message))

    def build_tree(self) -> ElementTree:
        root = Element("checkstyle", {"version": "4.3"})
        for filename in sorted(self.errors_by_file.keys()):
            errors = self.errors_by_file[filename]
            file = SubElement(root, "file", {"name": filename})
            for error in sorted(errors, key=lambda x: (x.symbol.line, x.symbol.column)):
                SubElement(
                    file,
                    "error",
                    {
                        "line": str(error.symbol.line),
                        "column": str(error.symbol.column),
                        "severity": "error",
                        "message": error.message,
                    },
                )
        return ElementTree(root)


def analyze_stubs(
    mypy_conf_path: str,
    base_stubs_path: str,
    reference_stubs_path: Optional[str] = None,
    expected_mismatches_path: Optional[str] = None,
    checkstyle_report: Optional[str] = None,
) -> bool:
    """
    Determine if the (presumably) handwritten stubs in base_stubs_path are correct;
    i.e. if they match the API of the modules that they are stubbing.

    For this they are compared to reference stubs, which by default are generated
    with mypy's stubgen tool. For each type mismatch (e.g. different function signature,
    missing class member) a message will be printed to stdout. The function will return
    False if any mismatches are found, unless they have been declared as expected.

    :param mypy_conf_path: path to mypy.ini
    :param base_stubs_path: path to the directory that contains the stubs to analyze
    :param reference_stubs_path: Path to the folder that contains the reference stubs.
        If not provided mypy's stubgen tool will be used to generate them.
    :param expected_mismatches_path: Path to JSON file that defines expected mismatches.
        Example:

        .. code-block:: json

            {
                "my.module.function": "mismatch",
                "another.module.Class": "not_found"
            }
    :param checkstyle_report: if this path is given, a xml report in checkstyle format
        will be written.
    :return: True if the stubs in base_stubs_path are considered correct
    """
    success = True
    failed_count = 0
    total_count = 0
    expected_count = 0

    try:
        mismatches, unused_mismatches = setup_expected_mismatches(
            expected_mismatches_path
        )
    except (JSONDecodeError, SchemaError) as ex:
        write_error(
            str(ex),
            linesep,
            CHECK_FILE_ERROR.format(file_path=expected_mismatches_path),
        )
        success = False

    if success:
        # Prevent overloaded function definitions from appearing multiple times
        stub_types_base_map = {
            symbol: path
            for (symbol, path) in get_stub_types(base_stubs_path, mypy_conf_path)
        }
        if reference_stubs_path:
            stub_types_reference = set(
                stub for stub, _ in get_stub_types(reference_stubs_path, mypy_conf_path)
            )
        else:
            stub_types_reference = set(
                stub for stub, _ in generate_stub_types(base_stubs_path, mypy_conf_path)
            )
        checkstyle_writer = CheckStyleWriter(stub_types_base_map)
        for res in compare(stub_types_base_map.keys(), stub_types_reference):
            total_count += 1
            evaluation_result = evaluate_compare_result(
                res,
                mismatches,
                unused_mismatches,
                expected_mismatches_path,
                loggers=[write_error, checkstyle_writer.collect_error],
            )
            if evaluation_result is EvaluationResult.FAILURE:
                failed_count += 1
            elif evaluation_result is EvaluationResult.EXPECTED_FAILURE:
                expected_count += 1
        success = failed_count == 0

        if checkstyle_report:
            checkstyle_tree = checkstyle_writer.build_tree()
            checkstyle_tree.write(
                checkstyle_report, encoding="UTF-8", xml_declaration=True
            )

        if unused_mismatches:
            success = False
            symbols = linesep.join([f" - {mm}" for mm in unused_mismatches])
            write_error(
                linesep,
                UNUSED_DEFINITION_ERROR.format(
                    symbols=symbols, file_path=expected_mismatches_path
                ),
            )

    ignore_message = IGNORE_MESSAGE.format(ignored=expected_count)

    if success:
        summary = SUCCESS_MESSAGE.format(total=total_count)
        print("", summary, (ignore_message if expected_count > 0 else ""), sep=linesep)
    else:
        summary = FAIL_MESSAGE.format(total=total_count, failed=failed_count)
        write_error(
            "", summary, (ignore_message if expected_count > 0 else ""), sep=linesep
        )
    return success


def main() -> None:
    args = parse_command_line()
    success = analyze_stubs(
        args.config,
        args.stubs_handwritten,
        args.reference,
        args.expected_mismatches,
        args.checkstyle_report,
    )
    sys.exit(0 if success else 1)


if __name__ == "__main__":
    main()



from os import path

# Always prefer setuptools over distutils
from setuptools import setup

VERSION = "0.1.3"
# This is the name of the GitHub repo and the package name on pypi
PACKAGE_NAME = "stubalyzer"


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    name=PACKAGE_NAME,
    version=VERSION,
    description=(
        "Analysis tool comparing hand written stubs to stubgen-generated ones, "
        "reporting inconsistencies"
    ),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url=f"https://github.com/kialo/{PACKAGE_NAME}",
    author="Kialo GmbH",
    author_email="open-source@kialo.com",
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Typing :: Typed",
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords="mypy analysis stubgen stubs",
    packages=["stubalyzer"],
    python_requires=">=3.6",
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["mypy", "schema"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={
        "dev": [
            "black",
            "flake8-rst-docstrings",
            "flake8",
            "ipython",
            "isort",
            "pip-tools",
            "pytest",
        ]
    },
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    entry_points={
        "console_scripts": ["stubalyzer=stubalyzer.analyze:main"]
    },  # Optional
    # List additional URLs that are relevant to your project as a dict.
    project_urls={
        "Bug Reports": f"https://github.com/kialo/{PACKAGE_NAME}/issues",
        "Source": f"https://github.com/kialo/{PACKAGE_NAME}",
    },
    # For mypy to find types
    zip_safe=False,
    package_data={"stubalyzer": ["py.typed"]},
)



"""
Compare mypy types.
"""

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, NamedTuple, Optional, Union

from mypy.nodes import (
    ARG_NAMED,
    ARG_POS,
    CONTRAVARIANT,
    COVARIANT,
    Decorator,
    FuncDef,
    SymbolNode,
    TypeAlias,
    TypeInfo,
    TypeVarExpr,
)
from mypy.subtypes import is_subtype
from mypy.types import CallableType, Overloaded
from mypy.types import Type as TypeNode

from .types import RelevantSymbolNode
from .utils import get_expression_fullname


class MatchResult(Enum):
    MATCH = "match"
    MISMATCH = "mismatch"
    NOT_FOUND = "not_found"
    MISLOCATED_SYMBOL = "mislocated_symbol"

    @classmethod
    def declare_mismatch(cls, matchResultString: str) -> MatchResult:
        err = matchResultString == MatchResult.MATCH.value

        try:
            result = MatchResult(matchResultString)
        except ValueError:
            err = True

        if err:
            possible_values = ", ".join(
                [f'"{m.value}"' for m in MatchResult if m is not MatchResult.MATCH]
            )
            raise ValueError(
                f'"{matchResultString}" is not a valid mismatch type.'
                f" (Use one of {possible_values}"
            )
        return result


def _get_symbol_type_info(symbol: SymbolNode) -> str:
    """
    Get the type of the given symbol as a human readable string.

    :param symbol: symbol for which to get the type
    """
    if isinstance(symbol, TypeAlias):
        return repr(symbol.target)
    if isinstance(symbol, TypeVarExpr):
        return _format_type_var(symbol)
    if isinstance(symbol, TypeInfo):
        return f"Class({symbol.fullname()})"

    return repr(getattr(symbol, "type", None))


class ComparisonResult(NamedTuple):
    """
    Result of comparing two symbol nodes and their types.
    """

    match_result: MatchResult
    """Type of comparison result"""

    symbol: RelevantSymbolNode
    """Symbol that was checked"""

    reference: Optional[SymbolNode]
    """Reference symbol that was checked against"""

    symbol_name: str
    """Full name of the symbol that was checked"""

    symbol_type: str
    """Type of the symbol that was checked"""

    reference_name: Optional[str]
    """Full name of the reference symbol"""

    reference_type: Optional[str]
    """Type of the reference symbol"""

    data: Optional[Dict[str, Any]] = None
    """Optional additional data"""

    message_val: Optional[str] = None
    """Optional message"""

    @property
    def message(self) -> str:
        """Human readable result of the comparison"""
        if self.message_val:
            return self.message_val

        if self.match_result is MatchResult.MATCH:
            return "\n".join(
                [
                    f"Types for {self.symbol_name} match:",
                    f"    {self.symbol_type}",
                    f"    {self.reference_type}",
                ]
            )
        elif self.match_result is MatchResult.MISMATCH:
            return "\n".join(
                [
                    f"Types for {self.symbol_name} do not match:",
                    f"    Handwritten type: {self.symbol_type}",
                    f"    Reference type  : {self.reference_type}",
                ]
            )
        elif self.match_result is MatchResult.NOT_FOUND:
            return f'Symbol "{self.symbol_name}" not found in generated stubs'
        elif self.match_result is MatchResult.MISLOCATED_SYMBOL:
            return (
                f'Found symbol "{self.symbol_name}" in different location'
                f' "{self.reference_name}".'
            )

    @classmethod
    def create(
        cls,
        match_result: MatchResult,
        symbol: RelevantSymbolNode,
        reference: Optional[SymbolNode],
        data: Optional[Dict[str, Any]] = None,
        message: Optional[str] = None,
    ) -> ComparisonResult:
        """
        Create a comparison result.

        :param match_result: if the match was successful
        :param symbol: symbol that was checked
        :param reference: reference symbol that was checked against
        :param data: optional additional data
        :param message: optional message
        """
        return cls(
            match_result=match_result,
            symbol=symbol,
            reference=reference,
            data=data,
            message_val=message,
            symbol_name=symbol.fullname(),
            symbol_type=_get_symbol_type_info(symbol),
            reference_name=reference.fullname() if reference else None,
            reference_type=_get_symbol_type_info(reference) if reference else None,
        )

    @classmethod
    def create_not_found(
        cls, symbol: RelevantSymbolNode, data: Optional[Dict[str, Any]] = None
    ) -> ComparisonResult:
        """
        Create an unsuccessful comparison result
        where there was no reference symbol found.

        :param symbol: symbol we wanted to check
        :param data: optional additional data
        """
        return cls.create(
            match_result=MatchResult.NOT_FOUND, symbol=symbol, reference=None, data=data
        )

    @classmethod
    def create_mislocated_symbol(
        cls,
        symbol: RelevantSymbolNode,
        reference: SymbolNode,
        data: Optional[Dict[str, Any]] = None,
    ) -> ComparisonResult:
        """
        Create an unsuccessful comparison result where the reference symbol was found
        in a different level of the class hierarchy.

        :param symbol: symbol we wanted to check
        :param reference: symbol that was found somewhere else in the hierarchy
        :param data: optional additional data
        """
        return cls.create(
            match_result=MatchResult.MISLOCATED_SYMBOL,
            symbol=symbol,
            reference=reference,
            data=data,
        )

    @classmethod
    def create_mismatch(
        cls,
        symbol: RelevantSymbolNode,
        reference: RelevantSymbolNode,
        data: Optional[Dict[str, Any]] = None,
        message: Optional[str] = None,
    ) -> ComparisonResult:
        """
        Create an unsuccessful comparison result.

        :param symbol: symbol that was checked
        :param reference: reference symbol that was checked against
        :param data: optional additional data
        :param message: optional message
        """
        return cls.create(
            match_result=MatchResult.MISMATCH,
            symbol=symbol,
            reference=reference,
            data=data,
            message=message,
        )

    @classmethod
    def create_match(
        cls,
        symbol: RelevantSymbolNode,
        reference: RelevantSymbolNode,
        data: Optional[Dict[str, Any]] = None,
        message: Optional[str] = None,
    ) -> ComparisonResult:
        """
        Create a successful comparison result.

        :param symbol: symbol that was checked
        :param reference: reference symbol that was checked against
        :param data: optional additional data
        :param message: optional message
        """
        return cls.create(
            match_result=MatchResult.MATCH,
            symbol=symbol,
            reference=reference,
            data=data,
            message=message,
        )


def _mypy_types_match(symbol_type: TypeNode, reference_type: TypeNode) -> MatchResult:
    """
    Check if the given symbol type matches the the reference type.

    :param symbol_type: symbol type to check
    :param reference_type: reference type to check against
    """
    if is_subtype(symbol_type, reference_type):
        return MatchResult.MATCH
    return MatchResult.MISMATCH


def _check_arguments_compatible(
    callable_type_or_func_def: Union[CallableType, FuncDef],
    reference: Union[CallableType, FuncDef],
) -> bool:
    """
    Check if argument kinds and names are compatible. This is used as a fallback, if
    mypy does not provide type information or is not strict enough.

    :param callable_type_or_func_def: CallableType or FuncDef to check arg_kinds and
        arg_names on
    :param reference: reference CallableType or FuncDef to check against
    """
    callable_args = list(
        zip(callable_type_or_func_def.arg_kinds, callable_type_or_func_def.arg_names)
    )
    reference_args = list(zip(reference.arg_kinds, reference.arg_names))

    positional_required_identical = [
        (kind, name) for (kind, name) in callable_args if kind == ARG_POS
    ] == [(kind, name) for (kind, name) in reference_args if kind == ARG_POS]

    keyword_required_equal = {
        (kind, name) for (kind, name) in callable_args if kind == ARG_NAMED
    } == {(kind, name) for (kind, name) in reference_args if kind == ARG_NAMED}

    optional_args = {
        (kind, name)
        for (kind, name) in callable_args
        if kind not in {ARG_POS, ARG_NAMED}
    }
    optional_reference_args = {
        (kind, name)
        for (kind, name) in reference_args
        if kind not in {ARG_POS, ARG_NAMED}
    }
    optional_compatible = optional_args.issubset(optional_reference_args)

    return (
        positional_required_identical and keyword_required_equal and optional_compatible
    )


def _callable_types_match(
    callable_type: CallableType, reference_callable: CallableType
) -> MatchResult:
    """
    Check if the given callable matches the reference.

    :param callable_type: callable to check
    :param reference_callable: callable to check against
    """
    arguments_compatible = _check_arguments_compatible(
        callable_type, reference_callable
    )
    if not arguments_compatible:
        return MatchResult.MISMATCH

    return _mypy_types_match(callable_type, reference_callable)


def _overloaded_types_match(
    overloaded: Overloaded, reference_overloaded: Overloaded
) -> MatchResult:
    """
    Check if the given overloaded type matches the reference.

    :param overloaded: overloaded type to check
    :param reference_overloaded: overloaded type to check against
    """
    if len(overloaded.items()) != len(reference_overloaded.items()):
        return MatchResult.MISMATCH

    for ovl, ref in zip(overloaded.items(), reference_overloaded.items()):
        if _callable_types_match(ovl, ref) != MatchResult.MATCH:
            return MatchResult.MISMATCH

    return MatchResult.MATCH


def compare_mypy_types(
    symbol: RelevantSymbolNode,
    reference: RelevantSymbolNode,
    symbol_type: Optional[TypeNode],
    reference_type: Optional[TypeNode],
) -> ComparisonResult:
    """
    Check if the mypy type of given symbol node is compatible with the reference
    symbol.

    Returns a successful comparison if:

    -  the reference type is None (this means mypy doesn't have enough information)
    -  the symbol type is a subtype of the reference type
    -  the symbol type overlaps with the reference type

    :param symbol: symbol node to validate
    :param reference: symbol node to validate against
    :param symbol_type: type of the symbol to validate
    :param reference_type: type of the symbol to validate against
    """
    if reference_type is None:
        # The reference type will never be None for overloaded functions.
        if isinstance(symbol, FuncDef):
            assert isinstance(reference, FuncDef)
            arguments_compatible = _check_arguments_compatible(symbol, reference)
            if arguments_compatible:
                return ComparisonResult.create_match(
                    symbol=symbol,
                    reference=reference,
                    message="Matched function definition without reference type.",
                )
            else:
                return ComparisonResult.create_mismatch(
                    symbol=symbol, reference=reference, message="Arguments don't match."
                )
        else:
            # MyPy does not have enough type information
            # hence we accept that our stub is correct
            return ComparisonResult.create_match(
                symbol=symbol, reference=reference, message="Generated type is None"
            )

    if symbol_type is None:
        return ComparisonResult.create_mismatch(symbol=symbol, reference=reference)

    if isinstance(symbol_type, CallableType) and isinstance(
        reference_type, CallableType
    ):
        match = _callable_types_match(symbol_type, reference_type)
    elif isinstance(symbol_type, Overloaded) and isinstance(reference_type, Overloaded):
        match = _overloaded_types_match(symbol_type, reference_type)
    else:
        match = _mypy_types_match(symbol_type, reference_type)

    return ComparisonResult.create(
        match_result=match, symbol=symbol, reference=reference
    )


def _type_infos_are_same_class(
    symbol: TypeInfo, reference: TypeInfo
) -> ComparisonResult:
    """
    Check if two TypeInfo symbols are the same class.

    This currently only does a comparison of the full name,
    since we only care if the classes are defined at the same location.
    The instance fields and methods are usually checked individually already.

    :param symbol: type info symbol to validate
    :param reference: type info symbol to validate against
    """
    if symbol.fullname() == reference.fullname():
        return ComparisonResult.create_match(symbol=symbol, reference=reference)
    else:
        return ComparisonResult.create_mismatch(symbol=symbol, reference=reference)


def _compare_type_aliases(symbol: TypeAlias, reference: TypeAlias) -> ComparisonResult:
    """
    Check if a TypeAlias symbol is a valid subtype of the given reference.

    This is done by comparing the target types of the aliases.

    :param symbol: type alias symbol to validate
    :param reference: type alias symbol to validate against
    """
    return compare_mypy_types(symbol, reference, symbol.target, reference.target)


def _format_type_var(symbol: TypeVarExpr) -> str:
    """
    Format a TypeVarExpr as it would be written in code.

    :param symbol: TypeVarExpr to format
    """

    variance = ""
    if symbol.variance == COVARIANT:
        variance = ", covariant=True"
    elif symbol.variance == CONTRAVARIANT:
        variance = ", contravariant=True"

    values = ""
    if symbol.values:
        values = ", " + (", ".join(str(t) for t in symbol.values))

    return f"{symbol.name()} = TypeVar('{symbol.name()}'{values}{variance})"


def _match_type_var_expr(symbol: TypeVarExpr, reference: TypeVarExpr) -> MatchResult:
    if symbol.variance != reference.variance:
        return MatchResult.MISMATCH

    if not symbol.values and not reference.values:
        return compare_mypy_types(
            symbol, reference, symbol.upper_bound, reference.upper_bound
        ).match_result

    if len(symbol.values) != len(reference.values):
        return MatchResult.MISMATCH

    for symbol_type, reference_type in zip(symbol.values, reference.values):
        result = compare_mypy_types(symbol, reference, symbol_type, reference_type)
        if result.match_result is not MatchResult.MATCH:
            return result.match_result

    return MatchResult.MATCH


def _compare_type_var_expr(
    symbol: TypeVarExpr, reference: TypeVarExpr
) -> ComparisonResult:
    """
    Check if a TypeVarExpr symbol matches the reference.

    :param symbol: type var symbol to validate
    :param reference: type var symbol to validate against
    """
    return ComparisonResult.create(
        match_result=_match_type_var_expr(symbol, reference),
        symbol=symbol,
        reference=reference,
    )


def _compare_decorator(symbol: Decorator, reference: Decorator) -> ComparisonResult:
    """
    Check if Decorator symbol matches the reference

    Returns a successful comparison if:
        - all decorators are the same and applied in the same order,
        - the function these decorators are applied to match

    :param symbol: decorator symbol to validate
    :param reference: decorator symbol to validate against
    """

    symbol_decorators = list(map(get_expression_fullname, symbol.original_decorators))
    reference_decorators = list(
        map(get_expression_fullname, reference.original_decorators)
    )

    if symbol_decorators == reference_decorators:
        function_comparison = compare_symbols(symbol.func, reference.func)
        return ComparisonResult.create(
            match_result=function_comparison.match_result,
            symbol=symbol,
            reference=reference,
            data=function_comparison.data,
            message=function_comparison.message,
        )
    else:
        return ComparisonResult.create_mismatch(
            symbol=symbol,
            reference=reference,
            data={
                "Symbol decorators": symbol_decorators,
                "Reference decorators": reference_decorators,
            },
            message=(
                f"Function {symbol.func.fullname()} stubs have different decorators."
            ),
        )


def compare_symbols(
    symbol: RelevantSymbolNode, reference: RelevantSymbolNode
) -> ComparisonResult:
    """
    Check if the given symbol node is compatible with the reference symbol.

    Will return a successful comparison if any of the following holds:

    -  the symbols describe the same class

    -  the symbols are type aliases that resolve to the same type

    -  ``symbol`` is a valid subtype of ``reference``
       (see :py:func:`mypy.subtypes.is_subtype`)

    -  ``symbol`` and ``reference`` somehow overlap
       (see :py:func:`mypy.meet.is_overlapping_types`)

    :param symbol: symbol node to validate
    :param reference: symbol node to validate against
    """
    # TODO: Check if this is always the case, i.e. could there be
    # cases where `symbol` and `reference` don't have the same class but still match?
    if type(symbol) != type(reference):
        return ComparisonResult.create_mismatch(symbol=symbol, reference=reference)

    if isinstance(symbol, TypeInfo) and isinstance(reference, TypeInfo):
        return _type_infos_are_same_class(symbol, reference)

    if isinstance(symbol, TypeAlias) and isinstance(reference, TypeAlias):
        return _compare_type_aliases(symbol, reference)

    if isinstance(symbol, TypeVarExpr) and isinstance(reference, TypeVarExpr):
        return _compare_type_var_expr(symbol, reference)

    if isinstance(symbol, Decorator) and isinstance(reference, Decorator):
        return _compare_decorator(symbol, reference)

    return compare_mypy_types(
        symbol, reference, getattr(symbol, "type"), getattr(reference, "type")
    )



from json.decoder import JSONDecodeError
from unittest.mock import Mock, patch

import pytest
from _pytest.capture import CaptureFixture
from mypy.nodes import TypeAlias, TypeVarExpr, Var
from schema import SchemaError

from .analyze import (
    EvaluationResult,
    analyze_stubs,
    compare,
    evaluate_compare_result,
    setup_expected_mismatches,
    write_error,
)
from .compare import ComparisonResult, MatchResult


class TestCompare:
    def test_skip_private_symbols(self) -> None:
        handwritten = [
            Mock(fullname=Mock(return_value="_private0"), spec=TypeAlias),
            Mock(fullname=Mock(return_value="_private1"), spec=TypeVarExpr),
            Mock(fullname=Mock(return_value="_private2"), spec=Var),
            Mock(fullname=Mock(return_value="public0")),
            Mock(fullname=Mock(return_value="__public1")),
        ]
        assert len(list(compare(handwritten, []))) == 2

    @patch(
        "stubalyzer.analyze.compare_symbols",
        return_value=ComparisonResult.create_match(Mock(), Mock()),
    )
    def test_match_to_generated_symbols(self, compare_mock: Mock) -> None:
        handwritten = [
            Mock(fullname=Mock(return_value="found_in_generated")),
            Mock(fullname=Mock(return_value="not_found_in_generated")),
        ]
        generated = [Mock(fullname=Mock(return_value="found_in_generated"))]
        result = list(compare(handwritten, generated))
        assert len(result) == 2
        assert result[0].match_result == MatchResult.MATCH
        assert result[1].match_result == MatchResult.NOT_FOUND


class TestSetupExpectedMismatches:
    def test_file_not_provided(self) -> None:
        assert setup_expected_mismatches(None) == ({}, set())

    def test_file_not_exists(self, capsys: CaptureFixture) -> None:
        assert setup_expected_mismatches("a_file") == ({}, set())
        _, err = capsys.readouterr()
        assert err.startswith("WARNING: Provided")

    @patch("pathlib.Path.read_text", return_value='{\'1: "4"}')
    @patch("pathlib.Path.exists", return_value=True)
    def test_wrong_json(self, exists_mock: Mock, read_mock: Mock) -> None:
        with pytest.raises(JSONDecodeError, match=r".*double quotes.*"):
            assert setup_expected_mismatches("a_file")

    @patch("pathlib.Path.read_text", return_value='{"lib.1": "match"}')
    @patch("pathlib.Path.exists", return_value=True)
    def test_json_contains_matches(self, *mocks: Mock) -> None:
        with pytest.raises(
            SchemaError, match=r'.*"match" is not a valid mismatch type.*'
        ):
            assert setup_expected_mismatches("a_file")

    @patch("pathlib.Path.read_text", return_value='{"lib.1": "unknown_match"}')
    @patch("pathlib.Path.exists", return_value=True)
    def test_json_contains_invalid_types(self, *mocks: Mock) -> None:
        with pytest.raises(
            SchemaError, match=r'.*"unknown_match" is not a valid mismatch type.*'
        ):
            assert setup_expected_mismatches("a_file")

    @patch("pathlib.Path.read_text", return_value="[1,2,3]")
    @patch("pathlib.Path.exists", return_value=True)
    def test_json_invalid_format(self, *mocks: Mock) -> None:
        with pytest.raises(SchemaError, match=r".*should be instance of 'dict'.*"):
            assert setup_expected_mismatches("a_file")

    @patch("pathlib.Path.read_text", return_value="{}")
    @patch("pathlib.Path.exists", return_value=True)
    def test_json_empty_dict(self, *mocks: Mock) -> None:
        assert setup_expected_mismatches("a_file") == ({}, set())

    @patch(
        "pathlib.Path.read_text",
        return_value='{"lib.1": "not_found", "lib.2": "mismatch"}',
    )
    @patch("pathlib.Path.exists", return_value=True)
    def test_json_is_correct(self, *mocks: Mock) -> None:
        assert setup_expected_mismatches("a_file") == (
            {"lib.1": MatchResult.NOT_FOUND, "lib.2": MatchResult.MISMATCH},
            {"lib.1", "lib.2"},
        )


class TestEvaluateCompareResult:
    def test_everything_ok(self, capsys: CaptureFixture) -> None:
        mismatches_left = {"1", "2", "4"}
        mismatches = {"1": MatchResult.NOT_FOUND}
        compare_result = Mock()
        compare_result.configure_mock(
            match_result=MatchResult.MATCH, message="Alright", symbol_name="3"
        )
        assert (
            evaluate_compare_result(
                compare_result, mismatches, mismatches_left, loggers=[write_error]
            )
            is EvaluationResult.SUCCESS
        )
        _, err = capsys.readouterr()
        assert err == ""
        assert mismatches_left == {"1", "2", "4"}

    def test_accept_expected_mismatches(self, capsys: CaptureFixture) -> None:
        mismatches_left = {"1", "2", "3", "4"}
        mismatches = {"1": MatchResult.NOT_FOUND, "3": MatchResult.MISMATCH}

        compare_result = Mock()
        compare_result.configure_mock(
            match_result=MatchResult.MISMATCH, symbol_name="3"
        )
        assert (
            evaluate_compare_result(
                compare_result, mismatches, mismatches_left, loggers=[write_error]
            )
            is EvaluationResult.EXPECTED_FAILURE
        )
        _, err = capsys.readouterr()
        assert err == ""
        assert mismatches_left == {"1", "2", "4"}

    def test_unwanted_mismatch(self, capsys: CaptureFixture) -> None:
        mismatches_left = {"1"}
        compare_result = Mock()
        compare_result.configure_mock(
            match_result=MatchResult.MISMATCH,
            message="An error happened",
            symbol_name="3",
        )
        assert (
            evaluate_compare_result(
                compare_result, {}, mismatches_left, loggers=[write_error]
            )
            is EvaluationResult.FAILURE
        )
        _, err = capsys.readouterr()
        assert "An error happened" in err
        assert mismatches_left == {"1"}

    def test_unwanted_match_instead_mismatch(self, capsys: CaptureFixture) -> None:
        mismatches_left = {"1", "2", "3", "4"}
        mismatches = {"1": MatchResult.NOT_FOUND, "3": MatchResult.MISMATCH}

        compare_result = Mock()
        compare_result.configure_mock(match_result=MatchResult.MATCH, symbol_name="3")

        expected_mismatches_path = "the/mismatch/path.json"
        assert (
            evaluate_compare_result(
                compare_result,
                mismatches,
                mismatches_left,
                expected_mismatches_path,
                loggers=[write_error],
            )
            is EvaluationResult.FAILURE
        )

        _, err = capsys.readouterr()
        assert 'Expected "3" to be "mismatch" but it matched.' in err
        assert mismatches_left == {"1", "2", "4"}

    def test_unwanted_match_instead_not_found(self, capsys: CaptureFixture) -> None:
        mismatches_left = {"1", "2", "3", "4"}
        mismatches = {"1": MatchResult.NOT_FOUND, "3": MatchResult.NOT_FOUND}

        compare_result = Mock()
        compare_result.configure_mock(match_result=MatchResult.MATCH, symbol_name="3")
        expected_mismatches_path = "the/mismatch/path.json"

        assert (
            evaluate_compare_result(
                compare_result,
                mismatches,
                mismatches_left,
                expected_mismatches_path,
                loggers=[write_error],
            )
            is EvaluationResult.FAILURE
        )

        _, err = capsys.readouterr()
        assert 'Expected "3" to be "not_found" but it matched.' in err
        assert mismatches_left == {"1", "2", "4"}

    def test_wrong_mismatch_type(self, capsys: CaptureFixture) -> None:
        mismatches_left = {"1", "2", "3", "4"}
        mismatches = {"1": MatchResult.NOT_FOUND, "3": MatchResult.MISMATCH}

        compare_result = Mock()
        compare_result.configure_mock(
            match_result=MatchResult.NOT_FOUND, symbol_name="3"
        )
        expected_mismatches_path = "the/mismatch/path.json"

        assert (
            evaluate_compare_result(
                compare_result,
                mismatches,
                mismatches_left,
                expected_mismatches_path,
                loggers=[write_error],
            )
            is EvaluationResult.FAILURE
        )

        _, err = capsys.readouterr()
        assert 'Expected "3" to be "mismatch" but it was "not_found".' in err
        assert mismatches_left == {"1", "2", "4"}


class TestAnalyzeStubs:
    @patch(
        "stubalyzer.analyze.evaluate_compare_result",
        return_value=EvaluationResult.SUCCESS,
    )
    @patch("stubalyzer.analyze.compare", return_value=range(10))
    @patch("stubalyzer.analyze.generate_stub_types")
    @patch("stubalyzer.analyze.get_stub_types", return_value=[])
    def test_everything_ok(
        self,
        generate_mock: Mock,
        compare_mock: Mock,
        result_mock: Mock,
        get_stub_types_mock: Mock,
        capsys: CaptureFixture,
    ) -> None:
        assert analyze_stubs("mypy_conf_path", "base_stubs_path")
        std, err = capsys.readouterr()
        assert "Successfully validated 10 stubs." in std
        assert (
            "0 fail(s) were ignored, "
            "because they were defined to be expected mismatches." not in std
        )
        assert "" == err

    @patch(
        "stubalyzer.analyze.setup_expected_mismatches",
        side_effect=JSONDecodeError("Boom", '{"a": 2}', 4),
    )
    def test_json_error(self, setup_mock: Mock, capsys: CaptureFixture) -> None:
        assert not analyze_stubs("mypy_conf_path", "base_stubs_path")
        _, err = capsys.readouterr()
        assert "Boom: line 1 column 5 (char 4)" in err

    @patch(
        "stubalyzer.analyze.setup_expected_mismatches", side_effect=SchemaError("Boom")
    )
    def test_schema_error(self, setup_mock: Mock, capsys: CaptureFixture) -> None:
        assert not analyze_stubs("mypy_conf_path", "base_stubs_path")
        _, err = capsys.readouterr()
        assert "Boom" in err

    @patch("stubalyzer.analyze.setup_expected_mismatches", return_value=({}, set()))
    @patch(
        "stubalyzer.analyze.evaluate_compare_result",
        side_effect=[EvaluationResult.FAILURE] * 2
        + [EvaluationResult.SUCCESS] * 4
        + [EvaluationResult.EXPECTED_FAILURE] * 4,
    )
    @patch("stubalyzer.analyze.compare", return_value=range(10))
    @patch("stubalyzer.analyze.generate_stub_types")
    @patch("stubalyzer.analyze.get_stub_types", return_value=[])
    def test_some_results_fail_ok(
        self,
        generate_mock: Mock,
        compare_mock: Mock,
        result_mock: Mock,
        setup_mock: Mock,
        get_stub_types_mock: Mock,
        capsys: CaptureFixture,
    ) -> None:
        assert not analyze_stubs(
            "mypy_conf_path",
            "base_stubs_path",
            expected_mismatches_path="a/proper/mismatch_path",
        )
        std, err = capsys.readouterr()
        assert "Failure: 2 of 10 stubs seem not to be valid." in err
        assert (
            "4 fail(s) were ignored, "
            "because they were defined to be expected mismatches." in err
        )
        assert "" == std

    @patch(
        "stubalyzer.analyze.setup_expected_mismatches",
        return_value=({}, ["lib.1", "lib.2"]),
    )
    @patch("stubalyzer.analyze.compare", return_value=[])
    @patch("stubalyzer.analyze.generate_stub_types")
    @patch("stubalyzer.analyze.get_stub_types", return_value=[])
    def test_unused_mismatches(
        self,
        generate_mock: Mock,
        compare_mock: Mock,
        setup_mock: Mock,
        get_stub_types_mock: Mock,
        capsys: CaptureFixture,
    ) -> None:
        assert not analyze_stubs(
            "mypy_conf_path",
            "base_stubs_path",
            expected_mismatches_path="a/proper/mismatch_path",
        )

        _, err = capsys.readouterr()
        assert (
            "Expected the following symbols to fail, but they were not processed:\n"
            " - lib.1\n"
            " - lib.2\n"
            'Check "a/proper/mismatch_path" to fix.' in err
        )
